import itertools, numpy as np, pandas as pd
from collections.abc import Iterable
from six import string_types
_numtypes = (int,float,np.generic)
_adj_admissable_types = (pd.Index, pd.Series, pd.DataFrame)

# Content:
# 0. Small auxiliary functions
# 1. cartesianProductIndex: Creates sparse, cartesian product from iterator of indices.
# 2. OrdSet: A small class that works like an ordered set.
# 3. adj: A small class used to subset and adjust pandas-like symbols.
# 4. adjMultiIndexDB, adjMultiIndex: A couple of classes that helps broadcasting pandas symbols defined over different indices.

### -------- 	0: Small, auxiliary functions    -------- ###
def tryint(x):
	try:
		return int(x)
	except ValueError:
		return x

def ifInt(x):
	try:
		int(x)
		return True
	except ValueError:
		return False

def return_version(x,dict_):
	if x not in dict_:
		return x
	elif (x+'_0') not in dict_:
		return x+'_0'
	else:
		maxInt = max([int(y.split('_')[-1]) for y in dict_ if (y.rsplit('_',1)[0]==x and ifInt(y.split('_')[-1]))])
		return x+'_'+str(maxInt+1)

def noneInit(x,FallBackVal):
	return FallBackVal if x is None else x

def dictInit(key,df_val,kwargs):
	return kwargs[key] if key in kwargs else df_val

def is_iterable(arg):
	return isinstance(arg, Iterable) and not isinstance(arg, string_types)

def getIndex(symbol):
	""" Defaults to None if no index is defined. """
	if hasattr(symbol, 'index'):
		return symbol.index
	elif isinstance(symbol, pd.Index):
		return symbol
	elif not is_iterable(symbol):
		return None

def getValues(symbol):
	""" Defaults to the index, if no values are defined (e.g. if symbol is an index) """
	if isinstance(symbol, (pd.Series, pd.DataFrame, pd.Index)):
		return symbol
	elif hasattr(symbol,'vals'):
		return symbol.vals
	elif not is_iterable(symbol):
		return symbol

def getDomains(x):
	return [] if getIndex(x) is None else getIndex(x).names

def domains_vlist(vlist):
	return OrdSet().union(*[OrdSet(getDomains(vi)) for vi in vlist]).v

### -------- 	1. Cartesian produt index     -------- ###
def cartesianProductIndex(indices):
	""" Return the cartesian product of pandas indices; assumes no overlap in levels of indices. """
	if any((i.empty for i in indices)):
		return pd.MultiIndex.from_tuples([], names = [n for l in indices for n in l.names]) 
	else: 
		ndarray = fastCartesianProduct([i.values for i in indices])
		return pd.MultiIndex.from_arrays(concatArrays(ndarray, indices).T, names = [n for l in indices for n in l.names])

# Auxiliary function for cartesianProductIndex
def fastCartesianProduct(arrays):
	la = len(arrays)
	L = *map(len, arrays), la
	dtype = np.result_type(*arrays)
	arr = np.empty(L, dtype=dtype)
	arrs = *itertools.accumulate(itertools.chain((arr,), itertools.repeat(0, la-1)), np.ndarray.__getitem__),
	idx = slice(None), *itertools.repeat(None, la-1)
	for i in range(la-1, 0, -1):
		arrs[i][..., i] = arrays[i][idx[:la-i]]
		arrs[i-1][1:] = arrs[i]
	arr[..., 0] = arrays[0][idx]
	return arr.reshape(-1, la)

# Auxiliary function for cartesianProductIndex
def getndarray(onedarray):
	return pd.MultiIndex.from_tuples(onedarray).to_frame(index=False).values

# Auxiliary function for cartesianProductIndex
def ndarray_or_1darray(ndarray, indices, i):
	return getndarray(ndarray[:,i]) if isinstance(indices[i], pd.MultiIndex) else ndarray[:,i:i+1]

# Auxiliary function for cartesianProductIndex
def concatArrays(ndarray, indices):
	return np.concatenate(tuple(ndarray_or_1darray(ndarray, indices, i) for i in range(len(indices))), axis=1)

class OrdSet:
	def __init__(self,i=None):
		self.v = list(dict.fromkeys(noneInit(i,[])))

	def __iter__(self):
		return iter(self.v)

	def __len__(self):
		return len(self.v)

	def __getitem__(self,item):
		return self.v[item]

	def __setitem__(self,item,value):
		self.v[item] = value

	def __add__(self,o):
		return OrdSet(self.v+list(o))

	def __sub__(self,o):
		return OrdSet([x for x in self.v if x not in o])

	def union(self,*args):
		return OrdSet(self.__add__([x for l in args for x in l]))

	def intersection(self,*args):
		return OrdSet([x for l in self.union(args) for x in l if x in self.v])

	def update(self,*args):
		self.v = self.union(*args).v

	def copy(self):
		return OrdSet(self.v.copy())

class adj:
	@staticmethod
	def rc_AdjPd(symbol, alias = None, lag = None):
		if isinstance(symbol, pd.Index):
			return adj.AdjAliasInd(adj.AdjLagInd(symbol, lag=lag), alias = alias)
		elif isinstance(symbol, pd.Series):
			return symbol.to_frame().set_index(adj.AdjAliasInd(adj.AdjLagInd(symbol.index, lag=lag), alias=alias),verify_integrity=False).iloc[:,0]
		elif isinstance(symbol, pd.DataFrame):
			return symbol.set_index(adj.AdjAliasInd(adj.AdjLagInd(symbol.index, lag=lag), alias=alias),verify_integrity=False)
		elif isinstance(symbol, _numtypes):
			return symbol
		else:
			raise TypeError(f"rc_AdjPd only uses instances {_adj_admissable_types} (and no scalars). Input was type {type(symbol)}")
	@staticmethod
	def AdjLagInd(index_,lag=None):
		if lag:
			if isinstance(index_,pd.MultiIndex):
				return index_.set_levels([index_.levels[index_.names.index(k)]+tryint(v) for k,v in lag.items()], level=lag.keys())
			elif list(index_.domains)==list(lag.keys()):
				return index_+list(lag.values())[0]
		else:
			return index_
	@staticmethod
	def AdjAliasInd(index_,alias=None):
		alias = noneInit(alias,{})
		return index_.set_names([x if x not in alias else alias[x] for x in index_.names])
	

	@staticmethod
	def rc_pd(s=None,c=None,alias=None,lag=None, pm = True, **kwargs):
		if isinstance(s,_numtypes):
			return s
		else:
			return adj.rctree_pd(s=s, c = c, alias = alias, lag = lag, pm = pm, **kwargs)
	@staticmethod
	def rc_pdInd(s=None,c=None,alias=None,lag=None,pm=True,**kwargs):
		if isinstance(s,_numtypes):
			return None
		else:
			return adj.rctree_pdInd(s=s,c=c,alias=alias,lag=lag,pm=pm,**kwargs)
	@staticmethod
	def rctree_pd(s=None,c=None,alias=None,lag =None, pm = True, **kwargs):
		a = adj.rc_AdjPd(s,alias=alias,lag=lag)
		if pm:
			return a[adj.point_pm(getIndex(a), c, pm)]
		else:
			return a[adj.point(getIndex(a) ,c)]
	@staticmethod
	def rctree_pdInd(s=None,c=None,alias=None,lag=None,pm=True,**kwargs):
		a = adj.rc_AdjPd(s,alias=alias,lag=lag)
		if pm:
			return getIndex(a)[adj.point_pm(getIndex(a), c, pm)]
		else:
			return getIndex(a)[adj.point(getIndex(a),c)]
	@staticmethod
	def point_pm(pdObj,vi,pm):
		if isinstance(vi,_adj_admissable_types):
			return adj.bool_ss_pm(pdObj,getIndex(vi),pm)
		elif isinstance(vi,dict):
			return adj.bool_ss_pm(pdObj,adj.rctree_pdInd(**vi),pm)
		elif isinstance(vi,tuple):
			return adj.rctree_tuple_pm(pdObj,vi,pm)
		elif vi is None:
			return pdObj == pdObj
	@staticmethod
	def point(pdObj, vi):
		if isinstance(vi, _adj_admissable_types):
			return adj.bool_ss(pdObj,getIndex(vi))
		elif isinstance(vi,dict):
			return adj.bool_ss(pdObj,adj.rctree_pdInd(**vi))
		elif isinstance(vi,tuple):
			return adj.rctree_tuple(pdObj,vi)
		elif vi is None:
			return pdObj == pdObj
	@staticmethod
	def rctree_tuple(pdObj,tup):
		if tup[0]=='not':
			return adj.translate_k2pd(adj.point(pdObj,tup[1]),tup[0])
		else:
			return adj.translate_k2pd([adj.point(pdObj,vi) for vi in tup[1]],tup[0])
	@staticmethod
	def rctree_tuple_pm(pdObj,tup,pm):
		if tup[0]=='not':
			return adj.translate_k2pd(adj.point_pm(pdObj,tup[1],pm),tup[0])
		else:
			return adj.translate_k2pd([adj.point_pm(pdObj,vi,pm) for vi in tup[1]],tup[0])
	@staticmethod
	def bool_ss(pdObjIndex,ssIndex):
		o,d = adj.overlap_drop(pdObjIndex,ssIndex)
		return pdObjIndex.isin([]) if len(o)<len(ssIndex.names) else pdObjIndex.droplevel(d).isin(adj.reorder(ssIndex,o))
	@staticmethod
	def bool_ss_pm(pdObjIndex,ssIndex,pm):
		o = adj.overlap_pm(pdObjIndex, ssIndex)
		if o:
			return pdObjIndex.droplevel([x for x in pdObjIndex.names if x not in o]).isin(adj.reorder(ssIndex.droplevel([x for x in ssIndex.names if x not in o]),o))
		else:
			return pdObjIndex==pdObjIndex if pm is True else pdObjIndex.isin([])
	@staticmethod
	def overlap_drop(pdObjIndex,index_):
		return [x for x in pdObjIndex.names if x in index_.names],[x for x in pdObjIndex.names if x not in index_.names]
	@staticmethod
	def overlap_pm(pdObjIndex,index_):
		return [x for x in pdObjIndex.names if x in index_.names]
	@staticmethod
	def reorder(index_,o):
		return index_ if len(index_.names)==1 else index_.reorder_levels(o)
	@staticmethod
	def translate_k2pd(l,k):
		if k == 'and':
			return sum(l)==len(l)
		elif k == 'or':
			return sum(l)>0
		elif k == 'not' and isinstance(l,(list,set)):
			return ~l[0]
		elif k == 'not':
			return ~l

### -------- 	4: Broadcasting methods    -------- ###
class adjMultiIndexDB:
	def __init__(self, db):
		self.db = db

	def bc(self,x,symbols,fill_value=0, sort_levels=None):
		v = self.sparsedomain([x]+symbols if is_iterable(symbols) else [x+symbols]).add(x,fill_value=fill_value).rename(x.name)
		return v if sort_levels is None else v.reorder_levels(sort_levels)

	def mergeDomains(self,symbols,c=None,sort_levels=None):
		v = self.sparsedomain(symbols, c = ('and', symbols) if c is None else c).dropna().index
		return v if sort_levels is None else v.reorder_levels(sort_levels)

	def sparsedomain(self, vlist, c=None):
		return pd.Series(0, index = adj.rc_pdInd(self.initindex_fromproduct(domains_vlist(vlist),self.db), c))
	
	def initindex_fromproduct(self, domains):
		return pd.MultiIndex.from_product([self.db.get(s) for s in domains]) if len(domains)>1 else self.db.get(domains[0])

class adjMultiIndex:
	@staticmethod
	def bc(x,y,fill_value = 0):
		""" Broadcast domain of 'x' to conform with domain of 'y'. """
		y, y_dom, x_dom = getIndex(y), getDomains(y), getDomains(x)
		if y_dom:
			if not x_dom:
				return pd.Series(x, index = y)
			elif set(x_dom).intersection(set(y_dom)):
				return x.add(pd.Series(0, index = y),fill_value=fill_value) if (set(x_dom)-set(y_dom)) else pd.Series(0, index = y).add(x,fill_value=fill_value)
			else:
				return pd.Series(0, index = cartesianProductIndex([getIndex(x),y])).add(x,fill_value=fill_value)
		else:
			return x
	@staticmethod
	def applyMult(symbol, mapping):
		""" Apply 'mapping' to a symbol using multiindex """
		if isinstance(symbol,pd.Index):
			return (pd.Series(0, index = symbol).add(pd.Series(0, index = adj.rc_pd(mapping,symbol)))).dropna().index.reorder_levels(symbol.names+[k for k in mapping.names if k not in symbol.names])
		elif isinstance(symbol,pd.Series):
			if symbol.empty:
				return pd.Series([], index = pd.MultiIndex.from_tuples([], names = symbol.index.names + [k for k in mapping.names if k not in symbol.index.names]))
			else: 
				return symbol.add(pd.Series(0, index = adj.rc_pd(mapping,symbol))).reorder_levels(symbol.index.names+[k for k in mapping.names if k not in symbol.index.names])
	@staticmethod
	def grid(v0,vT,index,gridtype='linear',phi=1):
		""" If v0, vT are 1d numpy arrays, returns 2d array. If scalars, returns 1d arrays. """
		if gridtype == 'linear':
			return np.linspace(v0,vT,len(index))
		elif gridtype=='polynomial':
			return np.array([v0+(vT-v0)*((i-1)/(len(index)-1))**phi for i in range(1,len(index)+1)])
	@staticmethod
	def addGrid(v0,vT,index,name,gridtype = 'linear', phi = 1,sort_levels=None, sort_index = False):
		""" NB: Make sure that v0 and vT are sorted similarly (if they are defined over indices, that is) """
		if sort_index:
			v0 = v0.sort_index()
			vT = vT.sort_index()
		if isinstance(v0,pd.Series):
			return pd.DataFrame(adjMultiIndex.grid(v0,vT,index,gridtype=gridtype,phi=phi).T, index = v0.index, columns = index).stack().rename(name).reorder_levels(index.names+v0.index.names if sort_levels is None else sort_levels)
		else:
			return pd.Series(adjMultiIndex.grid(v0,vT,index,gridtype=gridtype,phi=phi), index = index,name=name)
	


from pyDatabases.gpyDB.gpyDB import *
import openpyxl,io

# Content:
# 1. read: Small class of methods to load database from excel data.
# 2. robust: Robust methods to add/merge symbols and databases. 
# 3. adj: A class that adjusts the class 'adj' from _mixedTools.py.
# 4. aggregateDB: A class that aggregates GpyDB databases according to some mappings.

def sunion_empty(ls):
	""" return empty set if the list of sets (ls) is empty"""
	try:
		return set.union(*ls)
	except TypeError:
		return set()

def overlaps(s1,s2):
	doms1,doms2 = getDomains(s1), getDomains(s2)
	return [x for x in doms1 if x in doms2], [x for x in doms1 if x not in doms2], [x for x in doms2 if x not in doms1]

### -------- 	1: Read    -------- ###
class read:
	@staticmethod
	def SeriesDB_from_wb(workbook, kwargs, spliton='/'):
		""" 'read' should be a dictionary with keys = method, value = list of sheets to apply this to."""
		wb = simpleLoad(workbook) if isinstance(workbook,str) else workbook
		db = SeriesDB()
		[robust.merge_dbs_GpyDB(db,getattr(read, function)(wb[sheet],spliton=spliton)) for function, sheets in kwargs.items() for sheet in sheets];
		return db
	@staticmethod
	def simpleLoad(workbook):
		with open(workbook,"rb") as file:
			in_mem_file = io.BytesIO(file.read())
		return openpyxl.load_workbook(in_mem_file,read_only=True,data_only=True)
	@staticmethod
	def sheetnames_from_wb(wb):
		return (sheet.title for sheet in wb._sheets)
	@staticmethod
	def aux_map(sheet,col,spliton):
		pd_temp = sheet[col]
		pd_temp.columns = [x.split(spliton)[1] for x in pd_temp.iloc[0,:]]
		index = pd.MultiIndex.from_frame(pd_temp.dropna().iloc[1:,:])
		index.name = col
		return gpy(index,**{'name':col})
	@staticmethod
	def aux_var(sheet,col,spliton,type_):
		pd_temp = sheet[col].dropna()
		pd_temp.columns = [x.split(spliton)[1] for x in pd_temp.iloc[0,:]]
		if pd_temp.shape[1]==2:
			index = pd.Index(pd_temp.iloc[1:,0])
		else:
			index = pd.MultiIndex.from_frame(pd_temp.iloc[1:,:-1])
		return gpy(pd.Series(pd_temp.iloc[1:,-1].values,index=index,name=col),**{'type':type_})

	@staticmethod
	def sets(sheet, **kwargs):
		""" Return a dictionary with keys = set names and values = Gpy. na entries are removed. 
			The name of each set is defined as the first entry in each column. """
		pd_sheet = pd.DataFrame(sheet.values)
		return {pd_sheet.iloc[0,i]: gpy(pd.Index(pd_sheet.iloc[1:,i].dropna(),name=pd_sheet.iloc[0,i]),**{'name':pd_sheet.iloc[0,i]}) for i in range(pd_sheet.shape[1])}
	@staticmethod
	def subsets(sheet,spliton='/'):
		pd_sheet = pd.DataFrame(sheet.values)
		return {pd_sheet.iloc[0,i].split(spliton)[0]: gpy(pd.Index(pd_sheet.iloc[1:,i].dropna(),name=pd_sheet.iloc[0,i].split(spliton)[1]),**{'name': pd_sheet.iloc[0,i].split(spliton)[0]}) for i in range(pd_sheet.shape[1])}
	@staticmethod
	def maps(sheet,spliton='/'):
		pd_sheet = pd.DataFrame(sheet.values)
		pd_sheet.columns = [x.split(spliton)[0] for x in pd_sheet.iloc[0,:]]
		return {col: read.aux_map(pd_sheet,col,spliton) for col in set(pd_sheet.columns)}

	def variables(sheet,spliton='/',type_='variable'):
		pd_sheet = pd.DataFrame(sheet.values)
		pd_sheet.columns = [x.split(spliton)[0] for x in pd_sheet.iloc[0,:]]
		return {col: read.aux_var(pd_sheet,col,spliton,type_) for col in set(pd_sheet.columns)}
	
	def parameters(sheet,spliton='/'):
		return read.variables(sheet,spliton=spliton,type_='parameter')
	
	def scalar_variables(sheet,type_='variable',**kwargs):
		pd_sheet = pd.DataFrame(sheet.values)
		return {pd_sheet.iloc[i,0]: gpy(pd_sheet.iloc[i,1],**{'name':pd_sheet.iloc[i,0],'type':type_}) for i in range(pd_sheet.shape[0])}
	
	def scalar_parameters(sheet,**kwargs):
		return read.scalar_variables(sheet,type_='parameter')
	
	def variable2D(sheet,spliton='/',**kwargs):
		""" Read in 2d variable arranged in matrix; Note, only reads 1 variable per sheet."""
		pd_sheet = pd.DataFrame(sheet.values)
		domains = pd_sheet.iloc[0,0].split(spliton)
		var = pd.DataFrame(pd_sheet.iloc[1:,1:].values, index = pd.Index(pd_sheet.iloc[1:,0],name=domains[1]), columns = pd.Index(pd_sheet.iloc[0,1:], name = domains[2])).stack()
		var.name = domains[0]
		return {domains[0]: gpy(var,**kwargs)}
	

### -------- 	2: Merge databases    -------- ###
class robust:
	@staticmethod
	def robust_merge_dbs(db1,db2,priority=None):
		""" merge db2 into db1; priority = 'first' uses db1 if there is overlap. 'second' uses db2. This is much slower if priority = 'first'. """
		if isinstance(db1,(GpyDB,SeriesDB,dict)):
			if isinstance(db2,(GpyDB,SeriesDB,dict)):
				robust.merge_dbs_GpyDB(db1,db2,priority=priority)
			elif isinstance(db2,gams.GamsDatabase):
				robust.merge_dbs_GpyDB_gams(db1,db2,robust.get_g2np(db2),priority=priority)
		elif isinstance(db1,gams.GamsDatabase):
			if isinstance(db2,(GpyDB,SeriesDB,dict)):
				robust.merge_dbs_gams_GpyDB(db1,db2,robust.get_g2np(db1),priority=priority)
			elif isinstance(db2,gams.GamsDatabase):
				robust.merge_dbs_gams(db1,db2,robust.get_g2np(db1),priority=priority)
	@staticmethod
	def get_g2np(db):
		if isinstance(db,gams.GamsDatabase):
			return gams2numpy.Gams2Numpy(db.workspace.system_directory)
		elif isinstance(db,GpyDB):
			return db.g2np
		else:
			raise TypeError(f"db of type {type(db)} cannot access g2np.")
	@staticmethod
	def iters_db_py(db):
		return db if isinstance(db,(GpyDB,SeriesDB)) else db.values()
	@staticmethod
	def merge_dbs_GpyDB(db1,db2,priority=None):
		"""" merge db2 into db1. """
		if priority in ['second',None]:
			[GpyDBs_AOM_Second(db1,symbol) for symbol in robust.iters_db_py(db2)];
		elif priority== 'first':
			[GpyDBs_AOM_First(db1,symbol) for symbol in robust.iters_db_py(db2)];
	@staticmethod
	def merge_dbs_GpyDB_gams(db_py,db_gms,g2np,priority=None):
		""" Merge db_gms into db_py. """
		if priority in ['second',None]:
			[GpyDBs_AOM_Second(db_py,symbol) for symbol in dict_from_GamsDatabase(db_gms,g2np).values()];
		elif priority == 'first':
			[GpyDBs_AOM_First(db_py,symbol) for symbol in dict_from_GamsDatabase(db_gms,g2np).values()];
	@staticmethod
	def merge_dbs_gams_GpyDB(db_gms,db_py,g2np,priority=None):
		""" merge db_py into db_gms."""
		if priority in ['second',None]:
			[gpy2db_gams_AOM(s,db_gms,g2np,merge=True) for s in robust.iters_db_py(db_py)];
		elif priority == 'first':
			if isinstance(db_py,GpyDB):
				d = db_py.series.database.copy()
			elif isinstance(db_py,SeriesDB):
				d = db_py.symbols.copy()
			elif type(db_py) is dict:
				d = db_py.copy()
			robust.merge_dbs_GpyDB_gams(d, db_gms, g2np,priority='second') # merge db_gms into dictionary of gpy symbols.
			robust.merge_dbs_gams_GpyDB(db_gms, d, g2np,priority='second') # merge gpy symbols into gams.
	@staticmethod
	def merge_dbs_gams(db1,db2,g2np,priority=None):
		""" Merge db2 into db1. """
		if priority in ['second',None]:
			[gpy2db_gams_AOM(s,db1,g2np,merge=True) for s in dict_from_GamsDatabase(db2,g2np).values()];
		elif priority=='first':
			d = dict_from_GamsDatabase(db2,g2np) # copy of db2.
			robust.merge_dbs_GpyDB_gams(d,db1,g2np) # merge into d with priority to db1.
			robust.merge_dbs_gams_GpyDB(db1,d,g2np,priority='second') 

	@staticmethod	
	def robust_gpy(symbol,db=None,g2np=None,**kwargs):
		if isinstance(symbol,admissable_gpy_types):
			return gpy(symbol,**kwargs)
		elif isinstance(symbol,admissable_gamsTypes):
			return gpy(gpydict_from_GamsSymbol(db, g2np, symbol))
		else:
			try:
				return gpy(gpydict_from_GmdSymbol(db,g2np,symbol))
			except:
				raise TypeError(f"Tried to initiate gpy symbol from gams.Database._gmd. Check consistency of types: {type(symbol),type(db)}.")
	@staticmethod
	def robust_add(db,symbol,db_from=None,g2np=None,merge=False,**kwargs):
		""" Symbol ∈ {gams.database._GamsSymbol, pandas-like symbol, gpy}"""
		s = robust.robust_gpy(symbol,db=db_from, g2np = g2np, **kwargs)
		if isinstance(db,(dict, GpyDB, SeriesDB)):
			db[s.name] = s
		elif isinstance(db, gams.GamsDatabase):
			robust.gpy2db_gams(s,db[s.name],db,g2np,merge=merge)
		else:
			raise TypeError("Check type(db).")
	@staticmethod
	def robust_add_or_merge(db,symbol,db_from=None,g2np=None,merge=True,**kwargs):
		""" If 'symbol' exists in db merge with priority to new values in 'symbol'. """
		s = robust.robust_gpy(symbol,db=db_from, g2np = g2np, **kwargs)
		if isinstance(db,(dict, GpyDB,SeriesDB)):
			if s.name in symbols_db(db):
				db[s.name].vals = merge_gpy_vals(s.vals, db[s.name].vals)
			else:
				db[s.name] = s
		elif isinstance(db, gams.GamsDatabase):
			gpy2db_gams_AOM(s,db,g2np,merge=merge)
		else:
			raise TypeError("Check type(db).")
	
class adj(adj):
	@staticmethod
	def rc_AdjGpy(s, c = None, alias = None, lag = None, pm = True, **kwargs):
		if c is None:
			return adj.AdjGpy(s,alias=alias, lag = lag)
		else:
			copy = s.copy()
			copy.vals = adj.rc_pd(s=s,c=c,alias=alias,lag=lag,pm=pm)
			return copy
	@staticmethod
	def AdjGpy(symbol, alias = None, lag = None):
		copy = symbol.copy()
		copy.vals = adj.rc_AdjPd(symbol.vals, alias=alias, lag = lag)
		return copy

	@staticmethod
	def rc_AdjPd(symbol, alias = None, lag = None):
		if isinstance(symbol, pd.Index):
			return adj.AdjAliasInd(adj.AdjLagInd(symbol, lag=lag), alias = alias)
		elif isinstance(symbol, pd.Series):
			return symbol.to_frame().set_index(adj.AdjAliasInd(adj.AdjLagInd(symbol.index, lag=lag), alias=alias),verify_integrity=False).iloc[:,0]
		elif isinstance(symbol, pd.DataFrame):
			return symbol.set_index(adj.AdjAliasInd(adj.AdjLagInd(symbol.index, lag=lag), alias=alias),verify_integrity=False)
		elif isinstance(symbol,gpy):
			return adj.rc_AdjPd(symbol.vals, alias = alias, lag = lag)
		elif isinstance(symbol, (int,float,np.generic)):
			return symbol
		else:
			raise TypeError(f"Input was type {type(symbol)}")
	@staticmethod
	def rc_pd(s=None,c=None,alias=None,lag=None, pm = True, **kwargs):
		if isinstance(s,(int,float,np.generic)):
			return s
		elif isinstance(s, gpy) and (s.type in ('scalar_variable','scalar_parameter')):
			return s.vals
		else:
			return adj.rctree_pd(s=s, c = c, alias = alias, lag = lag, pm = pm, **kwargs)
	@staticmethod
	def rc_pdInd(s=None,c=None,alias=None,lag=None,pm=True,**kwargs):
		if isinstance(s,(int,float,np.generic)) or (isinstance(s,gpy) and (s.type in ('scalar_variable','scalar_parameter'))):
			return None
		else:
			return adj.rctree_pdInd(s=s,c=c,alias=alias,lag=lag,pm=pm,**kwargs)

class aggregateDB:
	# Main methods:
	# 1. updateSetValues: Update name of set elements in database using a mapping (dictionary)
	# 2. renameSet: Update name of a set in database.
	# 3. read_sets: Add sets to the database by reading established variables/parameters/mappings.
	# 4. update_sets: For existing database clean up set definitions and use 'read_sets' method to redefine sets.
	# 5. subset_db: Subset all symbols in database. 
	# 6. Aggregate database according to mapping. 
	def __init__(self,db):
		self.db = db

	def updateSetValues(self,set_,ns):
		""" Update set values for 'set_' using the namespace 'ns' """
		full_map = {k:k if k not in ns else ns[k] for k in self.db[set_]}
		for k,v in self.db.vardom(set_,types=['set','subset','mapping','parameter','variable']).items():
			[self.updateSetValue_Symbol(k,s,full_map) for s in v];
	
	def updateSetValue_Symbol(self,set_,s,ns):
		if not self.db.get(s).empty:
			if isinstance(self.db.get(s),pd.MultiIndex):
				self.db[s].vals = self.db.get(s).set_levels(self.db.get(s).levels[self.db[s].domains.index(set_)].map(ns),level=set_)
			elif isinstance(self.db.get(s),pd.Index):
				self.db[s].vals = self.db[s].index.map(ns).unique()
			elif isinstance(self.db[s].index,pd.MultiIndex):
				self.db.get(s).index = self.db[s].index.set_levels(self.db[s].index.levels[self.db[s].domains.index(set_)].map(ns),level=set_)
			elif isinstance(self.db[s].index,pd.Index):
				self.db.get(s).index = self.db[s].index.map(ns).unique()
	
	# ----------------------- 2. Rename set names ------------------------- #
	def renameSets(self, ns):
		""" 'ns' is a dictionary with key = original set, value = new set name. This does not alter aliases (unless they are included in 'ns') """
		[self.renameSet(self.db,k,v) for k,v in ns.items()];
	
	def renameSet(self,k,v):
		if k in self.db.symbols:
			self.db[v] = self.db.get(k).rename(v)
			self.db.series.__delitem__(k)
		[self.db.get(vi).__setattr__('index',self.db[vi].index.rename({k:v})) for vi in self.db.vardom(k,types=('variable','parameter'))[k]];
		[self.db.__setitem__(vi, self.db.get(vi).rename({k:v})) for vi in self.db.vardom(k,types=['mapping'])[k]];

	# ----------------------- 3-4. Read sets/update sets from database  ------------------------- #
	def read_sets(self, types=None, ignore_alias=False):
		""" read and define set elements from all symbols of type 'types'. """
		if ignore_alias:
			[add_or_merge_vals(self.db, symbol.index.get_level_values(set_).unique()) for symbol in self.db.getTypes(noneInit(types,['variable','parameter'])).values() for set_ in (set(symbol.domains)-self.db.alias_notin_db)];
		else:
			[add_or_merge_vals(self.db, symbol.index.get_level_values(set_).unique()) for symbol in self.db.getTypes(noneInit(types,['variable','parameter'])).values() for set_ in set(symbol.domains)];
	
	def update_sets(self, types = None, clean=True, ignore_alias=False, clean_alias = False):
		if clean:
			self.clean_sets()
		self.read_sets(types = types, ignore_alias = ignore_alias)
		if clean_alias:
			self.clean_aliases(types)
		self.read_aliased_sets(ignore_alias)
		if clean:
			self.update_subsets_from_sets()
			self.update_maps_from_sets()

	def clean_sets(self):
		""" create empty indices for all sets  """
		[self.db.__setitem__(set_, pd.Index([], name = set_)) for set_ in set(db.getTypes(['set']))-set(['alias_set','alias_map2'])];

	def clean_aliases(self,types):
		""" Remove aliases that are not used in variables/parameters """
		self.db.series['alias_'] = pd.MultiIndex.from_tuples(self.active_aliases(types), names = ['alias_set','alias_map2'])
		self.db.update_alias()

	def active_aliases(self,types):
		""" Return list of tuples with alias_ that are used in the model variables / mappings"""
		return [(k,v) for k in self.db.get('alias_set') for v in [x for x in self.db.alias_dict[k] if len(self.db.vardom(k,types=types)[x])>0]]
	
	def read_aliased_sets(self,ignore_alias):
		""" Read in all elements for aliased sets. If ignore alias"""
		for set_i in self.db.alias_dict:
			all_elements = sunion_empty([set(db.get(set_ij)) for set_ij in self.db.alias_dict0[set_i] if set_ij in self.db.getTypes(['set'])])
			if ignore_alias:
				[self.db.__setitem__(set_ij, pd.Index(all_elements,name=set_ij)) for set_ij in self.db.alias_dict0[set_i] if set_ij in self.db.getTypes(['set'])];
			else:
				[self.db.__setitem__(set_ij, pd.Index(all_elements,name=set_ij)) for set_ij in self.db.alias_dict0[set_i]];
	
	def update_subsets_from_sets(self):
		[self.update_subset(ss) for ss in self.db.getTypes(['subset'])];

	def update_subset(self,ss):
		if self.db.alias(self.db.get(ss).name) not in self.db.symbols:
			self.db.__setitem__(ss,pd.Index([],name=self.db.alias(self.db.get(ss).name)))
		else:
			self.db.__setitem__(ss,adj.rctree_pd(s=self.db[ss],c=self.db[self.db.alias(self.db.get(ss).name)]))
	
	def update_maps_from_sets(self):
		[self.update_map(m) for m in self.db.getTypes(['mapping'])];
	
	def update_map(self,m):
		if sum([bool(set(self.db.symbols.keys()).intersection(self.db.alias_all(s))) for s in self.db[m].domains])<len(self.db[m].domains):
			self.db.__setitem__(m, pd.MultiIndex.from_tuples([], names = self.db[m].domains))
		else:
			self.db.__setitem__(m, adj.rctree_pd(s=self.db[m], c = ('and', [self.db[s] for s in self.db[m].domains])))

	# ----------------------- 5. Subset database with index ------------------------- #
	def subset_db(self,index):
		[self.subset_db_valsFromList(index.rename(k), v) for k,v in self.db.vardom(index.name, types = ('set','subset','mapping','variable','parameter')).items()];

	def subset_db_valsFromList(self,index,listOfSymbols):
		[self.db[symbol].__setattr__('vals', adj.rctree_pd(self.db.get(symbol), index)) for symbol in listOfSymbols];

	# ----------------------- 6. Methods for aggregating database ------------------------- #
	def aggDB(self, mapping, aggBy=None, replaceWith=None, checkUnique=True, AggLike = None):
		""" Aggregate symbols in db according to mapping. This does so inplace, i.e. the set aggBy is altered. 
			Note: The aggregation assumes that mapping is 'one-to-many'; if this is not the case, a warning is printed (if checkUnique) """
		aggBy,replaceWith = noneInit(aggBy, mapping.names[0]), noneInit(replaceWith,mapping.names[-1])
		defaultAggLike = {k: ('Sum',{}) for v,l in self.db.vardom(aggBy).items() for k in l}
		AggLike = defaultAggLike if AggLike is None else defaultAggLike | AggLike
		[self.db.__setitem__(k, self.aggDB_set(k, mapping, aggBy, replaceWith)) for k in self.db.vardom(aggBy,types=['set'])];
		[self.db.__setitem__(vi, self.aggDB_subset(vi, mapping.set_names(k,level=aggBy), k, replaceWith, checkUnique)) for k,v in self.db.vardom(aggBy, types=['subset']).items() for vi in v];
		[self.db.__setitem__(vi, self.aggDB_mapping(vi, mapping.set_names(k,level=aggBy), k, replaceWith, checkUnique)) for k,v in self.db.vardom(aggBy, types=['mapping']).items() for vi in v];
		[self.db.__setitem__(vi, getattr(self,AggLike[vi][0])(self.db.get(vi),mapping.set_names(k,level=aggBy),k,replaceWith,checkUnique,**AggLike[vi][1])) for k,v in self.db.vardom(aggBy).items() for vi in v];
		return self.db

	def aggDB_set(self, k, mapping, aggBy, replaceWith):
		return mapping.get_level_values(replaceWith).unique().rename(k)
	
	def aggDB_subset(self, k, mapping, aggBy, replaceWith,checkUnique):
		o,d1,d2 = overlaps(self.db[k],mapping)
		if checkUnique:
			_checkUnique(self.db[k].index,mapping,o,d1,d2,aggBy,replaceWith,k)
		return self.aggReplace(self.db.get(k),mapping,aggBy,replaceWith,o).unique().rename(aggBy)
	
	def aggDB_mapping(self, k, mapping, aggBy, replaceWith, checkUnique):
		o,d1,d2 = overlaps(self.db[k],mapping)
		if checkUnique:
			self._checkUnique(self.db[k].index,mapping,o,d1,d2,aggBy,replaceWith,k)
		return self.aggReplace(self.db.get(k),mapping,aggBy,replaceWith,o).unique().set_names(aggBy,level=replaceWith)
	
	def aggVarSum(self, var, mapping, aggBy, replaceWith,checkUnique):
		o,d1,d2 = overlaps(var,mapping)
		if checkUnique:
			self._checkUnique(var.index,mapping,o,d1,d2,aggBy,replaceWith,var.name)
		return self.aggReplace(var,mapping,aggBy,replaceWith,o).rename_axis(index={replaceWith: aggBy}).groupby(var.index.names).sum().rename(var.name)
	
	def aggVarMean(self, var, mapping, aggBy, replaceWith,checkUnique):
		o,d1,d2 = overlaps(var,mapping)
		if checkUnique:
			self._checkUnique(var.index,mapping,o,d1,d2,aggBy,replaceWith,var.name)
		return self.aggReplace(var,mapping,aggBy,replaceWith,o).rename_axis(index={replaceWith: aggBy}).groupby(var.index.names).mean().rename(var.name)
	
	def aggVarSplitDistr(self,var,mapping,aggBy,replaceWith,checkUnique,weights=None):
		""" Can be used in many-to-one mappings to split up data with the key 'weights' """
		return (var*weights).dropna().droplevel(aggBy).rename_axis(index={replaceWith: aggBy}).reorder_levels(var.index.names).rename(var.name)
	
	def aggVarWeightedSum(self,var,mapping,aggBy,replaceWith,checkUnique,weights=None):
		return self.aggVarSum((var*weights).dropna().droplevel(replaceWith).reorder_levels(var.index.names),mapping,aggBy,replaceWith,checkUnique).rename(var.name)
	
	def aggVarWeightedSum_gb(self,var,mapping,aggBy,replaceWith,checkUnique,weights=None,sumOver=None):
		return self.aggVarWeightedSum(var,weights,mapping,aggBy,replaceWith,checkUnique).groupby([x for x in var.index.names if x not in sumOver]).sum()
	
	def aggVarLambda(self, var, mapping, aggBy, replaceWith,checkUnique, lambda_=None):
		o,d1,d2 = overlaps(var,mapping)
		if checkUnique:
			self._checkUnique(var.index,mapping,o,d1,d2,aggBy,replaceWith,var.name)
		return self.aggReplace(var,mapping,aggBy,replaceWith,o).rename_axis(index={replaceWith: aggBy}).groupby(var.index).sum(lambda_).rename(var.name)
	
	def _checkUnique(self, index1,index2,o,d1,d2,aggBy,replaceWith,name):
		mi1,mi2 = index1.to_frame().droplevel(d1), index2.reorder_levels(o+d2).to_frame().droplevel(d2)[d2]
		if d2:
			if max(rc_pd(mi2,mi1).groupby(mi2.index.names).nunique()[replaceWith])>1:
				print(f"""**** Warning: The symbol {name} used 'many-to-many' or 'many-to-one'-mapping. Aggregation usually assumes 'one-to-many'.""")
	
	def aggReplace(self,s,mapping,aggBy,replaceWith,overlap):
		return adjMultiIndex.applyMult(s,mapping.droplevel([v for v in mapping.names if v not in [replaceWith]+overlap])).droplevel(aggBy)



from ._database import *
admissable_gpy_types = (pd.Series,pd.Index,int,float,str,np.generic,dict,gpy)

# A few useful database functions
def dict_from_GamsDatabase(db_gams,g2np):
	return {symbol.name: gpy(gpydict_from_GamsSymbol(db_gams, g2np, symbol)) for symbol in db_gams if type(symbol) in admissable_gamsTypes}

def dict_from_GmdDatabase(db_gmd,g2np):
	rc = gmdcc.new_intp()
	syms = (gmdcc.gmdGetSymbolByNumberPy(db_gmd, i, rc) for i in range(gmdcc.gmdInfo(db_gmd, gmdcc.GMD_NRSYMBOLSWITHALIAS)[1]))
	return {NameFromGmd(db_gmd, s): gpy(gpydict_from_GmdSymbol(db_gmd,g2np,s)) for s in syms if IsGmdSymbolNotAliasEq(db_gmd, s)} | {'alias_': gpydict_Alias_from_Gmd(db_gmd, syms, rc)}

def gams_from_db_py(db_py,db_gams,g2np,merge=False):
	[gpy2db_gams_AOM(s,db_gams,g2np,merge=merge) for s in db_py];

class SeriesDB:
	""" Basically a wrapper that makes sure that iter works through self.database.values()"""
	def __init__(self,database=None):
		self.database = noneInit(database,{})

	def __iter__(self):
		return iter(self.database.values())

	def __len__(self):
		return len(self.database)

	def __getitem__(self,item):
		return self.database[item]

	def __setitem__(self,item,value):
		if item in self.database:
			if not is_iterable(value) and is_iterable(self[item].vals):
				value = pd.Series(value,index=self[item].index,name=self[item].name)
		self.database[item] = gpy(value,**{'name': item})

	def __delitem__(self,item):
		del(self.database[item])

	def get(self, item):
		return self.database[item].vals

	def copy(self):
		obj = type(self).__new__(self.__class__,None)
		obj.__dict__.update({k:v for k,v in self.__dict__.items() if k != 'database'})
		obj.database = {k:gpy(v) for k,v in self.database.items()}
		return obj

class GpyDB:
	def __init__(self,pickle_path=None,ws=None,db=None,alias=None,**kwargs):
		""" Init methods: 	(1) from file path to a pickle (fastest, only uses the kwargs 'ws','alias'), 
							(2) db = GpyDB (fast, only uses the kwarg 'ws','alias'), 
							(3) db = dict (fast, but does not check validity of attributes. only uses 'ws','alias'), 
							(4) db ∈ {None, str, gams.GamsDatabase} (slower, but more robust. uses all kwargs except 'pickle_path'); """ 
		if pickle_path is not None:
			self.init_from_pickle(pickle_path,ws=ws)
		elif isinstance(db,dict):
			self.update_with_ws(ws,db)
		elif isinstance(db,GpyDB):
			self.init_from_GpyGB(db,ws=ws)
		else:
			self.init_ws(ws)
			self.g2np = gams2numpy.Gams2Numpy(self.ws.system_directory)
			self.name = self.versionized_name(dictInit('name','rname',kwargs))
			self.export_settings = {'dropattrs': ['database','ws','g2np'], 'data_folder': dictInit('data_folder',os.getcwd(),kwargs)}
			self.init_dbs(db)
			self.series = SeriesDB(database=dict_from_GamsDatabase(self.database,self.g2np))
		self.update_alias(alias=alias)

	###################################################################################################
	###									0: Pickling/load/export settings 							###
	###################################################################################################
	def versionized_name(self,name):
		""" test if name is available in the current ws; update with an added if it is not """
		return return_version(name,self.ws._gams_databases)

	def init_from_pickle(self,pickle_path,ws=None):
		with open(pickle_path,"rb") as file:
			pickled_data=pickle.load(file)
		self.update_with_ws(ws,pickled_data.__dict__)

	def init_from_GpyGB(self,db,ws=None):
		self.update_with_ws(ws,db.__dict__)

	def update_with_ws(self,ws,dict_):
		if ws is None:
			self.__dict__ = dict_
		else:
			self.__dict__ = {key: value for key,value in dict_.items() if key not in ('ws','database')}
			self.init_ws(ws)
			self.name = self.versionized_name(self.name) # update name of database if it is already used in the current ws.
			self.init_dbs(dict_['database'])

	def init_ws(self,ws):
		if ws is None:
			self.ws = gams.GamsWorkspace()
		elif type(ws) is str:
			self.ws = gams.GamsWorkspace(working_directory=ws)
		elif isinstance(ws,gams.GamsWorkspace):
			self.ws = ws
		self.work_folder = self.ws.working_directory

	def init_dbs(self,db):
		if db is None:
			self.database = self.ws.add_database(database_name=self.name)
		elif type(db) is str:
			self.database = self.ws.add_database_from_gdx(db,database_name=self.name)
		elif isinstance(db,gams.GamsDatabase):
			self.database = self.ws.add_database(source_database=db,database_name=self.name)
		elif isinstance(db,Gpy_DB):
			self.database = self.ws.add_database(source_database=db.database,database_name=self.name)

	def update_alias(self,alias=None):
		alias = pd.MultiIndex.from_tuples([],names=['alias_set','alias_map2']) if alias is None else alias.set_names(['alias_set','alias_map2'])
		if 'alias_' not in self.series.database:
			self.series['alias_'] = alias
		else:
			self.series['alias_'] = self.get('alias_').union(alias)
		self.series['alias_set'] = self.get('alias_').get_level_values('alias_set').unique()
		self.series['alias_map2'] = self.get('alias_').get_level_values('alias_map2').unique()

	def __getstate__(self):
		if 'database' not in self.export_settings['dropattrs']:
			self.database.export(self.export_settings['data_folder']+'\\'+self.name+'.gdx')
		return {key: value for key,value in self.__dict__.items() if key not in (self.export_settings['dropattrs']+['database'])}

	def __setstate__(self,dict_):
		self.__dict__ = dict_
		try:
			self.ws = gams.GamsWorkspace(working_directory=self.work_folder)
		except FileNotFoundError:
			self.ws = gams.GamsWorkspace(working_directory=os.getcwd())
		self.g2np = gams2numpy.Gams2Numpy(self.ws.system_directory)
		if 'database' not in self.export_settings['dropattrs']:
			self.database = self.ws.add_database_from_gdx(self.export_settings['data_folder']+'\\'+self.name+'.gdx')
		else:
			self.database = self.ws.add_database()
		if 'series' in self.export_settings['dropattrs']:
			self.series = SeriesDB(database=dict_from_GamsDatabase(self.database,self.g2np))
	
	def export(self,name=None,repo=None):
		name = self.name if name is None else name
		repo = self.export_settings['data_folder'] if repo is None else repo
		with open(repo+'\\'+name, "wb") as file:
			pickle.dump(self,file)

	###################################################################################################
	###								1: Properties and base methods 									###
	###################################################################################################

	def __iter__(self):
		return self.series.__iter__()

	def __len__(self):
		return self.series.__len__()

	def __getitem__(self,item):
		try:
			return self.series[item]
		except KeyError:
			return self.series[self.alias(item)]

	def __setitem__(self,name,value):
		self.series.__setitem__(name,value)

	def get(self,item):
		try:
			return self.series[item].vals
		except KeyError:
			return self.series[self.alias(item)].vals.rename(item)

	@property
	def symbols(self):
		return self.series.database

	def getTypes(self,types):
		return {symbol.name: symbol for symbol in self.series if symbol.type in types}

	def copy(self,dropattrs=None,**kwargs):
		""" return copy of database. Ignore elements in dropattrs."""
		db = GpyDB(**{**self.__dict__,**kwargs})
		if 'series' not in noneInit(dropattrs,['database']) and 'series' not in kwargs.keys():
			db.series = self.series.copy()
		return db

	###################################################################################################
	###									2: Dealing with aliases			 							###
	###################################################################################################

	@property
	def alias_dict(self):
		return {name: self.get('alias_').get_level_values(1)[self.get('alias_').get_level_values(0)==name] for name in self.get('alias_set')}

	@property
	def alias_dict0(self):
		return {key: self.alias_dict[key].insert(0,key) for key in self.alias_dict}

	@property
	def alias_notin_db(self):
		return set(self.get('alias_map2'))-set(self.getTypes(['set']))

	def alias_all(self,x):
		if x in self.get('alias_set').union(self.get('alias_map2')):
			return self.alias_dict0[self.alias(x)]
		else: 
			return [x]

	def alias(self,x,index_=0):
		if x in self.get('alias_set'):
			return self.alias_dict0[x][index_]
		elif x in self.get('alias_map2'):
			key = self.get('alias_').get_level_values(0)[self.get('alias_').get_level_values(1)==x][0]
			return self.alias_dict0[key][index_]
		elif x in self.getTypes(['set','subset','mapping']) and index_==0:
			return x
		else:
			raise TypeError(f"{x} is not aliased")

	def domains_unique(self,x):
		""" Returns list of sets a symbol x is defined over. If x is defined over a set and its alias, only the set is returned. """
		return np.unique([self.alias(name) for name in self[x].index.names]).tolist()

	def vardom(self,set_,types=None):
		""" Returns a dict with keys = set_+aliases, values = list of symbols in 'types' defined over the the relevant set/alias"""
		return {set_i: [k for k,v in self.getTypes(noneInit(types,['variable','parameter'])).items() if set_i in v.domains] for set_i in self.alias_all(set_)}

	def merge_internal(self,merge=True):
		gams_from_db_py(self.series,self.database,self.g2np,merge=merge)





from pyDatabases._mixedTools import *
import os, gams, pickle, gams2numpy, gamstransfer, gmdcc

#global settings
default_attributes_variables = {k:i for k,i in zip(['level','marginal','lower','upper','scale'],range(-5,0))}
default_attributes_parameters= {k:i for k,i in zip(['value'],range(-1,0))}
admissable_py_types = (pd.Series,pd.Index,int,float,str,np.generic)
admissable_gamsTypes = (gams.GamsSet, gams.GamsParameter, gams.GamsVariable)


# Content:
# 0: Auxiliary functions used throughout
# 1: Methods used to identify the type of a symbol.
# 2: Methods used to write gams symbols to pandas representation (gpy_symbols).
#	 Includes methods for gams._GamsSymbol symbols and symbols from gams._gmd databases (Swig objects). 
# 3: From pandas/python symbol to gamstransfer.Container
# 4: From pandas/python symbol to database.
# 5. gpy: Class of user-defined version of symbols.
# 6: A few methods to merge symbols

### -------- 	0: Auxiliary functions    -------- ###
def symbols_db(db):
	""" return dictionary like version of database""" 
	return db if type(db) is dict else {symbol.name: symbol for symbol in db}

def tryint(x):
	try:
		return x.astype(int)
	except ValueError:
		return x

### -------- 	1: Identify types    -------- ###
def type_pandas_(symbol,name=None,gtype=None,**kwargs):
	if isinstance(symbol, pd.Series):
		return gtype if gtype else 'variable'
	elif isinstance(symbol,pd.MultiIndex):
		return 'mapping'
	elif isinstance(symbol,pd.Index):
		return set_or_subset(symbol.name,name)
	elif isinstance(symbol,(int,float,str,np.generic)):
		return 'scalar_parameter' if type in ('parameter','scalar_parameter') else 'scalar_variable'

def set_or_subset(s_name,name):
	return 'set' if s_name == name else 'subset'

def type_gamsSymbol_(symbol):
	if isinstance(symbol,gams.GamsVariable):
		return type_GamsVariable(symbol)
	elif isinstance(symbol,gams.GamsParameter):
		return type_GamsParameter(symbol)
	elif isinstance(symbol,gams.GamsSet):
		return type_GamsSet(symbol)
def type_GamsVariable(domains_as_strings):
	return 'scalar_variable' if not domains_as_strings else 'variable'
def type_GamsParameter(domains_as_strings):
	return 'scalar_parameter' if not domains_as_strings else 'parameter' 
def type_GamsSet(name, domains_as_strings):
	if len(domains_as_strings)==1:
		if domains_as_strings in [['*'], [name]]:
			return 'set'
		else:
			return 'subset'
	elif name!='SameAs':
		return 'mapping'

### -------- 	2: GAMS to pandas/gpy    -------- ###
# 2.1: From numpy like data to gpy-arranged dictionaries:
def np_from_gams(name,db_gams,g2np):
	return g2np.gmdReadSymbolStr(db_gams,name)

def df_from_variable(np_data,symbol_name,domains_as_strings,attributes=['level','marginal','lower','upper','scale']):
	return pd.DataFrame(np_data[:,[default_attributes_variables[k] for k in attributes]], index = index_from_np(np_data,symbol_name,domains_as_strings), columns = attributes)

def df_from_parameter(np_data,symbol_name,domains_as_strings,attributes=['value']):
	return pd.DataFrame(np_data[:,[default_attributes_parameters[k] for k in attributes]], index = index_from_np(np_data,symbol_name,domains_as_strings), columns = attributes)

def index_from_np(np_data,symbol_name,domains_as_strings):
	if len(domains_as_strings) > 1:
		return pd.MultiIndex.from_arrays([tryint(column) for column in np_data.T[0:len(domains_as_strings)]], names = index_names_from_np(symbol_name,domains_as_strings))
	elif len(domains_as_strings) == 1:
		return pd.Index(tryint(np_data[:,0]),name=index_names_from_np(symbol_name,domains_as_strings)[0])
	else:
		return None

def index_names_from_np(symbol_name,domains_as_strings):
	return [symbol_name] if domains_as_strings == ["*"] else domains_as_strings

def adjust_scalar(domains_as_strings,vals):
	return vals if domains_as_strings else vals[0]

def gpydict_from_Set(db_gams, g2np, name, domains_as_strings,type_,text=None):
	return {'vals': index_from_np(np_from_gams(name, db_gams, g2np), name, domains_as_strings), 'name': name, 'text': "" if text is None else text, 'type': type_}

def gpydict_from_Variable(db_gams, g2np, name, domains_as_strings,type_, text=None):
	return {'vals': adjust_scalar(domains_as_strings, df_from_variable(np_from_gams(name,db_gams,g2np), name, domains_as_strings,**{'attributes': ['level']})['level']), 'name': name, 'text': "" if text is None else text, 'type': type_}

def gpydict_from_Parameter(db_gams, g2np, name, domains_as_strings,type_, text=None):
	return {'vals': adjust_scalar(domains_as_strings, df_from_parameter(np_from_gams(name,db_gams,g2np), name, domains_as_strings)['value']), 'name': name, 'text': "" if text is None else text, 'type': type_}

# 2.2: From GamsDatabase._gmd symbols to gpy_symbol dictionaries:
def gpydict_from_GmdSymbol(db_gmd,g2np,symbol):
	if TypeFromGmdSymbol_NotAlias(db_gmd, symbol) == 'set':
		return gpydict_from_GmdSet(db_gmd,g2np,symbol)
	elif TypeFromGmdSymbol_NotAlias(db_gmd, symbol) == 'variable':
		return gpydict_from_GmdVariable(db_gmd, g2np, symbol)
	elif TypeFromGmdSymbol_NotAlias(db_gmd, symbol) == 'parameter':
		return gpydict_from_GmdParameter(db_gmd, g2np, symbol)

def gpydict_from_GmdSet(db_gmd,g2np,symbol):
	attrs = AttrFromGmd(db_gmd,symbol)
	return gpydict_from_Set(db_gmd, g2np, attrs['name'], attrs['domains_as_strings'], type_GamsSet(attrs['name'],attrs['domains_as_strings']), text=attrs['text'])

def gpydict_from_GmdVariable(db_gmd,g2np,symbol):
	attrs = AttrFromGmd(db_gmd,symbol)
	return gpydict_from_Variable(db_gmd, g2np, attrs['name'], attrs['domains_as_strings'], type_GamsVariable(attrs['domains_as_strings']), text=attrs['text'])

def gpydict_from_GmdParameter(db_gmd,g2np,symbol):
	attrs = AttrFromGmd(db_gmd,symbol)
	return gpydict_from_Parameter(db_gmd, g2np, attrs['name'], attrs['domains_as_strings'], type_GamsParameter(attrs['domains_as_strings']), text=attrs['text'])

def gpydict_Alias_from_Gmd(db_gmd, symbols, rc):
	return {'vals': AliasFromGmd(db_gmd,symbols,rc), 'name': 'alias_', 'type': 'mapping', 'text': 'Aliased sets'}

def AttrFromGmd(db_gmd,symbol):
	return {'name': NameFromGmd(db_gmd, symbol), 'domains_as_strings': gmdcc.gmdGetDomain(db_gmd,symbol, gmdcc.gmdSymbolInfo(db_gmd,symbol,gmdcc.GMD_DIM)[1])[-1],
	 'text': gmdcc.gmdSymbolInfo(db_gmd, symbol, gmdcc.GMD_EXPLTEXT)[-1], 'type': TypeFromGmdSymbol_NotAlias(db_gmd, symbol)}

def NameFromGmd(db_gmd, symbol):
	return gmdcc.gmdSymbolInfo(db_gmd, symbol, gmdcc.GMD_NAME)[-1]

def TypeFromGmdSymbol_NotAlias(db_gmd, symbol):
	return gamstransfer._Symbol._sym_object_map[(gmdcc.gmdSymbolType(db_gmd, symbol)[-1]), gmdcc.gmdSymbolInfo(db_gmd, symbol, gmdcc.GMD_USERINFO)[1]]

def IsGmdSymbolAlias(db_gmd,symbol):
	return gmdcc.gmdSymbolType(db_gmd, symbol)[-1]==gmdcc.GMS_DT_ALIAS

def IsGmdSymbolNotAliasEq(db_gmd,symbol):
	return gmdcc.gmdSymbolType(db_gmd,symbol)[-1] not in (gmdcc.GMS_DT_ALIAS,gmdcc.GMS_DT_EQU)

def AliasFromGmd(db_gmd,symbols,rc):
	return pd.MultiIndex.from_tuples([AliasTupleFromGmd(db_gmd,s,rc) for s in symbols if IsGmdSymbolAlias(db_gmd, s)],names=['alias_set','alias_map2'])

def AliasTupleFromGmd(db_gmd,symbol,rc):
	return (GetParentFromGmd(db_gmd, gmdcc.gmdSymbolInfo(db_gmd, symbol, gmdcc.GMD_USERINFO)[1],rc), NameFromGmd(db_gmd, symbol))

def GetParentFromGmd(db_gmd,subtype,rc):
	return gmdcc.gmdSymbolInfo(db_gmd, gmdcc.gmdGetSymbolByNumberPy(db_gmd,subtype-1,rc), gmdcc.GMD_NAME)[-1]

# 2.3: From gams._GamsSymbol to gpy_symbol dictionaries:
def gpydict_from_GamsSymbol(db_gams, g2np, symbol):
	if isinstance(symbol, gams.GamsSet):
		return gpydict_from_GamsSet(db_gams,g2np,symbol)
	elif isinstance(symbol, gams.GamsVariable):
		return gpydict_from_GamsVariable(db_gams,g2np,symbol)
	elif isinstance(symbol, gams.GamsParameter):
		return gpydict_from_GamsParameter(db_gams,g2np,symbol)

def gpydict_from_GamsSet(db_gams, g2np, symbol):
	return gpydict_from_Set(db_gams, g2np, symbol.name, symbol.domains_as_strings, type_GamsSet(symbol.name, symbol.domains_as_strings), symbol.text)

def gpydict_from_GamsVariable(db_gams, g2np, symbol):
	return gpydict_from_Variable(db_gams, g2np, symbol.name, symbol.domains_as_strings, type_GamsVariable(symbol.domains_as_strings), symbol.text)

def gpydict_from_GamsParameter(db_gams, g2np, symbol):
	return gpydict_from_Parameter(db_gams, g2np, symbol.name, symbol.domains_as_strings, type_GamsParameter(symbol.domains_as_strings), symbol.text)

### -------- 	3: Pandas/python to gams container    -------- ###
def gamstransfer_from_py_(symbol,container):
	""" symbol = gpy, container = gamstransfer.Container """
	if symbol.type == 'set':
		gamstransferSet_from_py(symbol.index, container, description=symbol.text)
	elif symbol.type == 'subset':
		gamstransferSubset_from_py(symbol.index, symbol.name, container, description = symbol.text)
	elif symbol.type == 'mapping':
		gamstransferMapping_from_py(symbol.index, symbol.name, container, description = symbol.text)
	elif symbol.type == 'scalar_variable':
		gamstransferScalarVariable_from_py(symbol.vals, symbol.name, container, description = symbol.text)
	elif symbol.type =='variable':
		gamstransferVariable_from_py(symbol.df, symbol.name, container, description=symbol.text)
	elif symbol.type == 'scalar_parameter':
		gamstransferScalarParameter_from_py(symbol.vals, symbol.name, container, description=symbol.text)
	elif symbol.type == 'parameter':
		gamstransferParameter_from_py(symbol.df, symbol.name, container, description=symbol.text)

def gamstransferSet_from_py(index,container,description=None):
	gamstransfer.Set(container, index.name, description="" if description is None else description, records = index.astype(str))

def gamstransferSubset_from_py(index,name,container,description=None):
	gamstransfer.Set(container, name, index.name, description="" if description is None else description, records = index.astype(str))

def gamstransferMapping_from_py(index,name,container,description=None):
	gamstransfer.Set(container, name, index.names, description="" if description is None else description, records = index.to_frame(index=False).astype(str))

def gamstransferScalarVariable_from_py(scalar, name, container, description=None):
	if isinstance(scalar,pd.DataFrame):
		gamstransfer.Variable(container, name, description = "" if description is None else description, records = scalar)
	elif not is_iterable(scalar):
		gamstransfer.Variable(container, name, description = "" if description is None else description, records = pd.DataFrame([scalar], columns = ['level']))

def gamstransferVariable_from_py(df, name, container, description = None):
	gamstransfer.Variable(container, name, domain= df.index.names, description="" if description is None else description, records = df.reset_index().astype({k:str for k in df.index.names}))

def gamstransferScalarParameter_from_py(scalar, name, container, description=None):
	gamstransfer.Parameter(container, name, description="", records = scalar)

def gamstransferParameter_from_py(df, name, container, description= None):
	gamstransfer.Parameter(container, name, domain = df.index.names, description="" if description is None else description, records = df.reset_index().astype({k:str for k in df.index.names}))

### -------- 	4: Pandas/python to database    -------- ###
def gpy2db_gams_AOM(s,db,g2np,merge=True):
	if s.name in gamsdb_symbols(db):
		gpy2db_gams(s,db[s.name],db,g2np,merge=merge)
	else:
		AddSymbol2db_gams(s,db,g2np)

def gamsdb_symbols(db):
	return {s.name: s for s in db}

def gpy2db_gams(s,symbolPtr,db,g2np,merge=False):
	if s.type in ('set','subset','mapping'):
		g2np.gmdFillSymbolStr(db,symbolPtr,set2np(s),merge=merge)
	elif s.type == 'pscalar':
		g2np.gmdFillSymbolStr(db,symbolPtr,pscalar2np(s),merge=merge)
	elif s.type == 'parameter':
		g2np.gmdFillSymbolStr(db,symbolPtr,parameter2np(s),merge=merge)
	elif s.type == 'scalar_variable':
		g2np.gmdFillSymbolStr(db,symbolPtr,scalar2np(s),merge=merge)
	elif s.type == 'variable':
		g2np.gmdFillSymbolStr(db,symbolPtr,variable2np(s),merge=merge)

def AddSymbol2db_gams(s,db,g2np):
	if s.type == 'set':
		g2np.gmdFillSymbolStr(db,db.add_set(s.name,1,s.text),set2np(s),merge=False)
	elif s.type in ('subset','mapping'):
		g2np.gmdFillSymbolStr(db,db.add_set_dc(s.name,s.domains,s.text),set2np(s),merge=False)
	elif s.type == 'scalar_parameter':
		g2np.gmdFillSymbolStr(db,db.add_parameter(s.name,0,s.text),pscalar2np(s),merge=False)
	elif s.type == 'parameter':
		g2np.gmdFillSymbolStr(db,db.add_parameter_dc(s.name,s.domains,s.text),parameter2np(s),merge=False)
	elif s.type == 'scalar_variable':
		g2np.gmdFillSymbolStr(db,db.add_variable(s.name,0,gams.VarType.Free,s.text),scalar2np(s),merge=False)
	elif s.type == 'variable':
		g2np.gmdFillSymbolStr(db,db.add_variable_dc(s.name,gams.VarType.Free,s.domains,s.text),variable2np(s),merge=False)

def AdjIndex(df):
	return df.reset_index().astype({k: str for k in df.index.names})
def set2np(s):
	return s.vals.to_frame(index=False).assign(_rname="").astype(str).to_numpy()
def pscalar2np(s):
	return np.array([[s.vals]])
def parameter2np(s):
	return AdjIndex(s.vals).to_numpy()
def scalar2np(s):
	return np.array([list((gamstransfer.Variable._default_values['free'] | {'level': s.vals}).values())])
def variable2np(s):
	return AdjIndex(s.df.assign(**{k:v for k,v in gamstransfer.Variable._default_values['free'].items() if k!='level'})).to_numpy()


### -------- 	5. gpy: class of symbols    -------- ###
class gpy:
	""" Customized class of symbols used in the PM/GPM database classes. """
	def __init__(self,symbol,**kwargs):
		""" Initialize with gpy, dict, or pandas object. Fast init with suitable dict/gpy. """
		if isinstance(symbol,(gpy,dict)):
			[setattr(self,key,value) for key,value in symbol.items() if key not in kwargs];
			[setattr(self,key,value) for key,value in kwargs.items()];
		elif isinstance(symbol,admissable_py_types):
			self.vals = symbol
			self.name = kwargs['name'] if 'name' in kwargs else symbol.name
			self.type = type_pandas_(symbol, **kwargs)
			self.text = dictInit('text',"",kwargs)

	def __iter__(self):
		return iter(self.vals)

	def __len__(self):
		return len(self.vals)

	def items(self):
		return self.__dict__.items()

	def copy(self):
		obj = type(self).__new__(self.__class__,None)
		obj.__dict__.update(self.__dict__)
		return obj

	@property
	def index(self):
		if isinstance(self.vals,pd.Index):
			return self.vals
		elif hasattr(self.vals,'index'):
			return self.vals.index
		else:
			return None
	@property
	def domains(self):
		return [] if self.index is None else self.index.names

	@property
	def df(self):
		""" Only works if self.vals is a pandas series + type is either variable/parameter."""
		return self.vals.to_frame(name='level' if self.type == 'variable' else 'value')


### -------- 	6: Add or merge symbols    -------- ###
def merge_gpy_vals(s1,s2):
	if isinstance(s1,pd.Series):
		return s1.combine_first(s2)
	elif isinstance(s1,pd.Index):
		return s1.union(s2)
	elif type_pandas_(s1) in ['scalar_variable','scalar_parameter']:
		return s1

def GpyDBs_AOM_Second(db,symbol):
	if symbol.name in symbols_db(db):
		db[symbol.name].vals = merge_gpy_vals(symbol.vals, db[symbol.name].vals)
	else:
		db[symbol.name] = symbol
def GpyDBs_AOM_First(db,symbol):
	if symbol.name in symbols_db(db):
		db[symbol.name].vals = merge_gpy_vals(db[symbol.name].vals,symbol.vals)
	else:
		db[symbol.name] = symbol

def add_or_merge_vals(db,symbol,name=None):
	if name is None:
		name = symbol.name
	if name in symbols_db(db):
		db[name].vals = merge_gpy_vals(symbol, db[name].vals)
	else:
		db[name] = symbol



from pyDatabases._mixedTools import *
from pyDatabases.simpleDB.simpleDB import SimpleDB, type_
import openpyxl,io

# 1: Read methods:
class read:
	@staticmethod
	def dbFromWB(workbook, kwargs, spliton='/'):
		""" 'read' should be a dictionary with keys = method, value = list of sheets to apply this to."""
		wb = read.simpleLoad(workbook) if isinstance(workbook,str) else workbook
		db = SimpleDB()
		[db.mergeDbs(getattr(read, function)(wb[sheet],spliton=spliton)) for function,sheets in kwargs.items() for sheet in sheets];
		return db

	@staticmethod
	def simpleLoad(workbook):
		with open(workbook,"rb") as file:
			in_mem_file = io.BytesIO(file.read())
		return openpyxl.load_workbook(in_mem_file,read_only=True,data_only=True)

	def sheetnames_from_wb(self, wb):
		return (sheet.title for sheet in wb._sheets)

	@staticmethod
	def sets(sheet, **kwargs):
		""" Return a dictionary with keys = set names and values = pandas objects. na entries are removed. 
			The name of each set is defined as the first entry in each column. """
		pd_sheet = pd.DataFrame(sheet.values)
		return {pd_sheet.iloc[0,i]: pd.Index(pd_sheet.iloc[1:,i].dropna(),name=pd_sheet.iloc[0,i]) for i in range(pd_sheet.shape[1])}

	@staticmethod
	def subsets(sheet,spliton='/'):
		pd_sheet = pd.DataFrame(sheet.values)
		return {pd_sheet.iloc[0,i].split(spliton)[0]: pd.Index(pd_sheet.iloc[1:,i].dropna(),name=pd_sheet.iloc[0,i].split(spliton)[1]) for i in range(pd_sheet.shape[1])}

	@staticmethod
	def aux_map(sheet,col,spliton):
		pd_temp = sheet[col]
		pd_temp.columns = [x.split(spliton)[1] for x in pd_temp.iloc[0,:]]
		return pd.MultiIndex.from_frame(pd_temp.dropna().iloc[1:,:])

	@staticmethod
	def maps(sheet,spliton='/'):
		pd_sheet = pd.DataFrame(sheet.values)
		pd_sheet.columns = [x.split(spliton)[0] for x in pd_sheet.iloc[0,:]]
		return {col: read.aux_map(pd_sheet,col,spliton) for col in set(pd_sheet.columns)}

	@staticmethod
	def aux_var(sheet,col,spliton):
		pd_temp = sheet[col].dropna()
		pd_temp.columns = [x.split(spliton)[1] for x in pd_temp.iloc[0,:]]
		if pd_temp.shape[1]==2:
			index = pd.Index(pd_temp.iloc[1:,0])
		else:
			index = pd.MultiIndex.from_frame(pd_temp.iloc[1:,:-1])
		return pd.Series(pd_temp.iloc[1:,-1].values,index=index,name=col)

	@staticmethod
	def variables(sheet,spliton='/'):
		pd_sheet = pd.DataFrame(sheet.values)
		pd_sheet.columns = [x.split(spliton)[0] for x in pd_sheet.iloc[0,:]]
		return {col: read.aux_var(pd_sheet,col,spliton) for col in set(pd_sheet.columns)}

	@staticmethod
	def scalars(sheet,**kwargs):
		pd_sheet = pd.DataFrame(sheet.values)
		return {pd_sheet.iloc[i,0]: pd_sheet.iloc[i,1] for i in range(pd_sheet.shape[0])}

	@staticmethod
	def variable2D(sheet,spliton='/',**kwargs):
		""" Read in 2d variable arranged in matrix; Note, only reads 1 variable per sheet."""
		pd_sheet = pd.DataFrame(sheet.values)
		domains = pd_sheet.iloc[0,0].split(spliton)
		var = pd.DataFrame(pd_sheet.iloc[1:,1:].values, index = pd.Index(pd_sheet.iloc[1:,0],name=domains[1]), columns = pd.Index(pd_sheet.iloc[0,1:], name = domains[2])).stack()
		var.name = domains[0]
		return {domains[0]: var}

# 2: 
def readSets(db, types = None):
	""" Read sets from database symbols """
	[db.addOrMerge(set_, getIndex(symbol).get_level_values(set_).unique()) for symbol in db.getTypes(noneInit(types,['variable'])).values() for set_ in getIndex(symbol).names];

# 3: Broadcasting-like methods
def applyMult(symbol, mapping):
	""" Apply 'mapping' to a symbol using multiindex """
	if isinstance(symbol,pd.Index):
		return (pd.Series(0, index = symbol).add(pd.Series(0, index = rc_pd(mapping,symbol)))).dropna().index.reorder_levels(symbol.names+[k for k in mapping.names if k not in symbol.names])
	elif isinstance(symbol,pd.Series):
		if symbol.empty:
			return pd.Series([], index = pd.MultiIndex.from_tuples([], names = symbol.index.names + [k for k in mapping.names if k not in symbol.index.names]))
		else: 
			return symbol.add(pd.Series(0, index = rc_pd(mapping,symbol))).reorder_levels(symbol.index.names+[k for k in mapping.names if k not in symbol.index.names])

def appendIndexWithCopy(index, copyLevel, newLevel):
	if is_iterable(copyLevel):
		return pd.MultiIndex.from_frame(index.to_frame(index=False).assign(**{newLevel[i]: index.get_level_values(copyLevel[i]) for i in range(len(copyLevel))}))
	else: 
		return pd.MultiIndex.from_frame(index.to_frame(index=False).assign(**{newLevel: index.get_level_values(copyLevel)}))

def broadcast(x,y,fill_value=0):
	""" y is a index or None, x is a scalar or series."""
	if type_(y) == 'set':
		if getDomains(y):
			if not getDomains(x):
				return pd.Series(x, index = y)
			elif set(getDomains(x)).intersection(set(getDomains(y))):
				if set(getDomains(x))-set(getDomains(y)):
					return x.add(pd.Series(0, index = y), fill_value=fill_value)
				else:
					return pd.Series(0, index=y).add(x,fill_value=fill_value)
			else:
				return pd.Series(0, index = cartesianProductIndex([database.getIndex(x),y])).add(x,fill_value=fill_value)
		else:
			return x
	else:
		b = broadcast(x, getIndex(y),fill_value=fill_value)
		return b.add(y,fill_value=fill_value) if isinstance(b,pd.Series) else x+y




import setuptools

with open("README.md", "r") as file:
  long_description = file.read()

setuptools.setup(
  name="pyDatabases",
  version="0.0.2",
  author="Rasmus K. Skjødt Berg",
  author_email="rasmus.kehlet.berg@econ.ku.dk",
  description="Small collection of database classes based primarily pandas, secondarily scipy and GAMS.",
  long_description=long_description,
  long_description_content_type="text/markdown",
  url="https://github.com/ChampionApe/pyDatabases",
  packages=setuptools.find_packages(),
  classifiers=[
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
  ],
  python_requires='>=3.8',
  install_requires=["pandas", "scipy","openpyxl"],
)


from pyDatabases._mixedTools import *
from pyDatabases._mixedTools import _numtypes
from copy import deepcopy

def type_(s):
	if isinstance(s, pd.Index):
		return 'set'
	elif isinstance(s, pd.Series):
		return 'variable'
	elif isinstance(s,_numtypes):
		return 'scalar'
	else:
		return 'other'
def mergeVals(s1,s2):
	if isinstance(s1,pd.Series):
		return s1.combine_first(s2)
	elif isinstance(s1,pd.Index):
		return s1.union(s2)
	else:
		return s1
def symbols_(db_i):
	""" return dictionary of symbols """
	return db_i.symbols if isinstance(db_i, SimpleDB) else db_i

class SimpleDB:
	""" Collection of data """
	def __init__(self,name='name',symbols=None,alias=None):
		self.name = name
		self.symbols = noneInit(symbols,{})
		self.updateAlias(alias=alias)

	def updateAlias(self,alias=None):
		self.alias = self.alias.union(pd.MultiIndex.from_tuples(noneInit(alias,[]), names = ['from','to'])) if hasattr(self,'alias') else pd.MultiIndex.from_tuples(noneInit(alias,[]), names = ['from','to'])

	def __iter__(self):
		return iter(self.symbols.values())

	def __len__(self):
		return len(self.symbols)

	def __getitem__(self,item):
		try:
			return self.symbols[item]
		except KeyError:
			try:
				return self.symbols[self.getAlias(item)].rename(item)
			except TypeError:
				raise TypeError(f"Symbol {item} not in database")

	def __setitem__(self,item,value):
		if item in self.symbols:
			if not is_iterable(value) and is_iterable(self[item]):
				value = pd.Series(value,index=self[item].index,name=self[item].name)
		self.symbols[item] = value

	def __delitem__(self,item):
		del(self.symbols[item])

	def copy(self):
		obj = type(self).__new__(self.__class__,None)
		obj.__dict__.update(deepcopy(self.__dict__).items())
		return obj

	def getTypes(self,types=['variable']):
		return {k:v for k,v in self.symbols.items() if type_(v) in types}

	def variableDomains(self,set_,types=['variable']):
		""" Return 'types' defined over 'set_'"""
		return {k:v for k,v in self.getTypes(types).items() if set_ in getDomains(v)}

	@property
	def aliasDict(self):
		return {k: self.alias.get_level_values(1)[self.alias.get_level_values(0) == k] for k in self.alias.get_level_values(0).unique()}

	@property
	def aliasDict0(self):
		return {key: self.aliasDict[key].insert(0,key) for key in self.aliasDict}

	def getAlias(self,x,index_=0):
		if x in self.alias.get_level_values(0):
			return self.aliasDict0[x][index_]
		elif x in self.alias.get_level_values(1):
			return self.aliasDict0[self.alias.get_level_values(0)[self.alias.get_level_values(1)==x][0]][index_]
		elif x in self.getTypes(['set']) and index_==0:
			return x
		else:
			raise TypeError(f"{x} is not aliased")

	def addOrMerge(self, name, symbol, priority = 'first'):
		if name in self.symbols:
			self[name] = mergeVals(self.symbols[name],symbol) if priority == 'first' else mergeVals(symbol, self.symbols[name])
		else:
			self[name] = symbol

	def mergeDbs(self, dbOther, priority='first'):
		""" Merge all symbols in two databases """
		[self.addOrMerge(name, symbol, priority=priority) for name,symbol in symbols_(dbOther).items()];
